/*

	IDA Loader module for Nintendo DS (NDS) ROM files.
	Written by Dennis Elser.

	Comments:
	---------
	This is the first loader module for IDA
	I have written so far. It might therefore
	give inaccurate results.
	Feedback, bugreports and donations are welcome =)

	Hint: actually having Nintendo DS hardware would
		  speed things up ;)
	
	dennis[at]backtrace[dot]de / www[dot]backtrace[dot]de


	Credits:
	--------
	-	Rafael Vuijk for the CRC16 Code and the
		NDS ROM Header layout
	-	DataRescue for providing loader sourcecode
		with the IDA 4.7 SDK

	History:
	--------

	2004/12/29 initial version
	
	- can disassemble arm7 or arm9 code
	- can add additional information about
	  the ROM to database

	2005/01/01 version 1.1

	- creates a section for the whole RAM area
	  and maps the cartridge's code into it
	- adds more additional information about
	  the ROM to database

*/


#include "../idaldr.h"
#include "nds.h"


//defines
#define version "v1.1"

//global data
nds_hdr hdr;

//--------------------------------------------------------------------------
//	this code has been taken from Rafael Vuijk's "ndstool"
//	and slightly been modified by me
//
unsigned short CalcCRC16(nds_hdr *ndshdr)
{
	unsigned short crc16 = 0xFFFF;
	for (int i=0; i<0x015E; i++)
	{
		unsigned char data = *((unsigned char *)ndshdr + i);
		crc16 = (crc16 >> 8) ^ crc16tab[(crc16 ^ data) & 0xFF];
	}
	return crc16;
}

//--------------------------------------------------------------------------
//
//      check input file format. if recognized, then return 1
//      and fill 'fileformatname'.
//      otherwise return 0
//
int accept_file(FILE *fp, char fileformatname[MAX_FILE_FORMAT_NAME], int n)
{

	ulong filelen;

	if(n)
		return 0;

	//get filesize
	filelen = efilelength(fp);

	//quit, if file is smaller than the size of an NDS header
	if (filelen < sizeof(nds_hdr))
		return 0;

	//set filepos to offset 0
	qfseek(fp, 0, SEEK_SET);

	//read whole NDS header
	if(qfread(fp, &hdr, sizeof(nds_hdr)) != sizeof(nds_hdr))
		return 0;

	//check validity of CRC16 value of header
	//this is used to determine if this file is an NDS file
	if (CalcCRC16(&hdr) != hdr.header_crc)
		return 0;

	//this is the name of the file format
	qstrncpy(fileformatname, "Nintendo DS ROM (NDS)", MAX_FILE_FORMAT_NAME);

	return (1 | ACCEPT_FIRST);
}



//--------------------------------------------------------------------------
//
//      load file into the database.
//
void load_file(FILE *fp, ushort /*neflag*/, const char * /*fileformatname*/)
{

	ea_t startEA;
	ea_t endEA;
	long offset;
	char code;

	//check if the processor is set to ARM
	if ( ph.id != PLFM_ARM )
	{
		//msg("Setting processor type to ARM710a.\n");
		set_processor_type("ARM710a", SETPROC_ALL|SETPROC_FATAL);
	}

	//go to file-offset 0
	qfseek(fp, 0, SEEK_SET);

	//and read the whole header
	qfread(fp, &hdr, sizeof(nds_hdr));

	//display a messagebox asking the user for details
	// 1 - Yes
	// 0 - No
	//-1 - Cancel
	int answer = askyn_cv(1,
		"NDS Loader by Dennis Elser.\n\n"
		"This file possibly contains ARM7 and ARM9 code.\n"
		"Choose \"Yes\" to load the ARM9 executable,\n"
		"\"No\" to load the ARM7 executable\n\n"
		"Do you want to load the ARM9 code?\n"
		
		,NULL
		);

	//user chose "cancel" ?
	if(answer==BADADDR)
		qexit(1);

	//user chose "yes" = arm9
	if(answer)
	{
		//init
		inf.startIP = inf.beginEA = hdr.arm9_entry_address;
		startEA = hdr.arm9_ram_address;
		endEA = hdr.arm9_ram_address+hdr.arm9_size;
		offset = hdr.arm9_rom_offset;
		code = '9';
		//sanitycheck
		if (efilelength(fp) < offset+hdr.arm9_size)
			qexit(1);
	}
	//user chose "no" = arm7
	else
	{
		//init
		inf.startIP = inf.beginEA = hdr.arm7_entry_address;
		startEA = hdr.arm7_ram_address;
		endEA = hdr.arm7_ram_address+hdr.arm7_size;
		offset = hdr.arm7_rom_offset;
		code = '7';
		//sanitycheck
		if (efilelength(fp) < offset+hdr.arm7_size)
			qexit(1);
	}
	
	//check if segment lies within RAM
	if(startEA < RAM_BEGIN || endEA > RAM_END)
		qexit(1);

	//map selector
	set_selector(1, 0);
	inf.start_cs = 1;

	//create a segment for the RAM
	if (!add_segm(1, RAM_BEGIN/*startEA*/, RAM_END/*endEA*/, "RAM", CLASS_CODE)) qexit(1);

	//enable 32bit addressing
	set_segm_addressing(getseg( startEA ), 1);

	//load file into RAM area
	file2base(fp, offset, startEA, endEA, FILEREG_PATCHABLE);
	
	//add additional information about the ROM to the database
	describe(RAM_BEGIN,true,"   NDS Loader %s by Dennis Elser.\n",version);
	describe(RAM_BEGIN,true,"   Game Title:         %s",hdr.title);
	describe(RAM_BEGIN,true,"   Code:               ARM%c",code);
	describe(RAM_BEGIN,true,"   ROM content:        0x%08X",startEA);
	describe(RAM_BEGIN,true,"   End of ROM content: 0x%08X\n",endEA);

	describe(startEA,true,"   Beginning of ROM content",NULL);
	describe(endEA,true,  "   End of ROM content",NULL);

}

//----------------------------------------------------------------------
//
//      LOADER DESCRIPTION BLOCK
//
//----------------------------------------------------------------------
loader_t LDSC =
{
  IDP_INTERFACE_VERSION,
  0,                            // loader flags
//
//      check input file format. if recognized, then return 1
//      and fill 'fileformatname'.
//      otherwise return 0
//
  accept_file,
//
//      load file into the database.
//
  load_file,
//
//      create output file from the database.
//      this function may be absent.
//
  NULL,
};
